"""fprime_gds.plugin.definitions: definitions of plugin specifications and decorators

In order to define a plugin, an implementation decorator is used. Users can import `gds_plugin_implementation` from this
file to decorate functions that implement plugins.

This file also defines helper classes to support the plugin system.

@author lestarch
"""

import inspect
import pluggy
from enum import Enum, auto
from typing import Any, Dict, Tuple, Type

PROJECT_NAME = "fprime_gds"

gds_plugin_specification = pluggy.HookspecMarker(PROJECT_NAME)
gds_plugin_implementation = pluggy.HookimplMarker(PROJECT_NAME)


def gds_plugin(plugin_class):
    """Decorator: make decorated class a plugin of the plugin_class

    This allows users to quickly define plugin implementations by decorating their class with the definition of the
    plugin they are implementing. This works by searching the plugin_class for a function with the attribute
    { PROJECT_NAME }_spec. This will be the specification function. It will then add an implementation of the
    implementation spec to the supplied class.

    It will check the following assertions:
      1. plugin_class has a valid gds_plugin_specification
      2. plugin_class specification is well-formed (no arguments, class method)
      2. The decorated class is a valid subclass of plugin_class
      3. The decorated class is non-abstract
    """
    assert isinstance(
        plugin_class, type
    ), "Must supply @gds_plugin valid plugin classes"
    plugin_name = plugin_class.__name__
    spec_attr = f"{ PROJECT_NAME }_spec"
    spec_methods = inspect.getmembers(
        plugin_class, predicate=lambda item: hasattr(item, spec_attr)
    )
    assert len(spec_methods) == 1, f"'{plugin_name}' is not a valid F Prime GDS plugin."
    spec_method = spec_methods[0]
    spec_method_name, spec_method_function = spec_method
    assert (
        getattr(spec_method_function, "__self__", None) == plugin_class
    ), f"'{plugin_name}' specification invalid."
    assert inspect.isabstract(
        plugin_class
    ), f"{plugin_class} is not a plugin superclass."

    def decorator(decorated_class):
        """Implementation of the decorator: check valid subclass, and add plugin method"""
        assert isinstance(decorated_class, type), "Must use @gds_plugin on classes"
        class_name = decorated_class.__name__
        assert issubclass(
            decorated_class, plugin_class
        ), f"{class_name} is not a subclass of {plugin_name}"
        assert not inspect.isabstract(
            decorated_class
        ), f"{class_name} is abstract. Plugins may not be abstract."

        def return_decorated_class(cls):
            """Function to become the plugin implementation method"""
            assert cls == decorated_class, "Plugin system failure"
            return decorated_class

        setattr(
            decorated_class,
            spec_method_name,
            classmethod(gds_plugin_implementation(return_decorated_class)),
        )
        return decorated_class

    return decorator


class PluginType(Enum):
    """Enumeration of plugin types"""

    ALL = auto()
    """ Plugin selection including all types of plugins """

    SELECTION = auto()
    """ Plugin that provides a selection between implementations """

    FEATURE = auto()
    """ Plugin that provides a feature """


class Plugin(object):
    """Plugin wrapper object"""

    def __init__(self, category: str, plugin_type: PluginType, plugin_class: Type[Any]):
        """Initialize the plugin

        Args:
            category: category of the plugin (i.e. register_<category>_function)
            plugin_type: type of plugin
            plugin_class: implementation class of the plugin
        """
        self.category = category
        self.type = plugin_type
        self.plugin_class = plugin_class

    def get_implementor(self):
        """Get the implementor of this plugin"""
        return self.plugin_class

    def get_name(self):
        """Get the name of the plugin

        Plugin names are derived from the `get_name` class method of the plugin's implementation class. When not defined
        that name is derived from the plugin's implementation class __name__ property instead.

        Returns:
            name of plugin
        """
        return (
            self.plugin_class.get_name()
            if hasattr(self.plugin_class, "get_name")
            else self.plugin_class.__name__
        )

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Get arguments needed by plugin

        Plugin argument are derived from the `get_arguments` class method of the plugin's implementation class. When not
        defined an empty dictionary is returned.

        Returns:
            argument specification for plugin
        """
        return (
            self.plugin_class.get_arguments()
            if hasattr(self.plugin_class, "get_arguments")
            else {}
        )

    def check_arguments(self, **kwargs):
        """Check a plugin's arguments

        Check a plugin's arguments if it defines a check method. Arguments are passed as kwargs just like the arguments are
        passed to the constructor.
        """
        if hasattr(self.plugin_class, "check_arguments"):
            self.plugin_class.check_arguments(**kwargs)



"""
cli.py:

This file sets up the command line interface and argument parsing that is done to support the F prime executable tools
layer. It is designed to allow users to import standard sets of arguments that applied to the various aspects of the
code that they are importing.

@author mstarch
"""

import argparse
import datetime
import errno
import functools
import getpass
import inspect
import itertools
import os
import platform
import re
import sys

import yaml

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Tuple

# Required to set the checksum as a module variable
import fprime_gds.common.communication.checksum
import fprime_gds.common.logger
from fprime_gds.common.communication.adapters.ip import check_port
from fprime_gds.common.pipeline.standard import StandardPipeline
from fprime_gds.common.transport import ThreadedTCPSocketClient
from fprime_gds.common.utils.config_manager import ConfigManager
from fprime_gds.executables.utils import find_app, find_dict, get_artifacts_root
from fprime_gds.plugin.definitions import PluginType
from fprime_gds.plugin.system import Plugins, PluginsNotLoadedException
from fprime_gds.common.zmq_transport import ZmqClient


GUIS = ["none", "html"]


class ParserBase(ABC):
    """Base parser for handling fprime command lines

    Parsers must define several functions. They must define "get_parser", which will produce a parser to parse the
    arguments, and an optional "handle_arguments" function to do any necessary processing of the arguments. Note: when
    handling arguments.
    """

    DESCRIPTION = None

    @property
    def description(self):
        """Return parser description"""
        return self.DESCRIPTION if self.DESCRIPTION else "Unknown command line parser"

    @abstractmethod
    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return argument list handled by this parser

        Produce the arguments that can be processed by multiple parsers. i.e. argparse, and pytest parsers are the
        intended consumers. Returns a tuple of dictionary of flag tuples (--flag, -f) to keyword arguments to pass to
        argparse and list of arguments calculated by the parser (generated).

        Returns:
            tuple of dictionary of flag tuple to keyword arguments, list of generated fields
        """

    def get_parser(self) -> argparse.ArgumentParser:
        """Return an argument parser to parse arguments here-in

        Produce a parser that will handle the given arguments. These parsers can be combined for a CLI for a tool by
        assembling them as parent processors to a parser for the given tool.

        Return:
            argparse parser for supplied arguments
        """
        parser = argparse.ArgumentParser(
            description=self.description,
            add_help=True,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        self.fill_parser(parser)
        return parser

    @staticmethod
    def safe_add_argument(parser, *flags, **keywords):
        """Add an argument allowing duplicates

        Add arguments to the parser (passes through *flags and **keywords) to the supplied parser. This method traps
        errors to prevent duplicates from crashing the system when two plugins use the same flags.

        Args:
            parser: parser or argument group to add arguments to
            *flags: positional arguments passed to `add_argument`
            **keywords: key word arguments passed to `add_argument`
        """
        try:
            parser.add_argument(*flags, **keywords)
        except argparse.ArgumentError:
            # flag has already been added, pass
            pass

    @classmethod
    def add_arguments_from_specification(cls, parser, arguments):
        """Safely add arguments to parser

        In parsers and plugins, arguments are represented as a map of flag tuples to argparse keyword arguments. This
        function will add arguments of that representation supplied as `arguments` to the supplied parser in a safe
        collision-avoidant manner.

        Args:
            parser: argparse Parser or ArgumentGroup, or anything with an `add_argument` function
            arguments: arguments specification

        """
        for flags, keywords in arguments.items():
            cls.safe_add_argument(parser, *flags, **keywords)

    def fill_parser(self, parser):
        """Fill supplied parser with arguments

        Fills the supplied parser with the arguments returned via the `get_arguments` method invocation. This
        implementation adds the arguments directly to the parser.

        Args:
            parser: parser to fill with arguments

        """
        self.add_arguments_from_specification(parser, self.get_arguments())

    def reproduce_cli_args(self, args_ns):
        """Reproduce the list of arguments needed on the command line"""

        def flag_member(flags, argparse_inputs) -> Tuple[str, str]:
            """Get the best CLI flag and namespace member"""
            best_flag = (
                [flag for flag in flags if flag.startswith("--")] + list(flags)
            )[0]
            member = argparse_inputs.get(
                "dest", re.sub(r"^-+", "", best_flag).replace("-", "_")
            )
            return best_flag, member

        def cli_arguments(flags, argparse_inputs) -> List[str]:
            """Get CLI argument list fro argument entry"""
            best_flag, member = flag_member(flags, argparse_inputs)
            value = getattr(args_ns, member, None)

            action = argparse_inputs.get("action", "store")
            assert action in [
                "store",
                "store_true",
                "store_false",
            ], f"{action} not supported by reproduce_cli_args"

            # Handle arguments
            if (action == "store_true" and value) or (
                action == "store_false" and not value
            ):
                return [best_flag]
            elif action != "store" or value is None:
                return []
            return [best_flag] + (
                [str(item) for item in value]
                if isinstance(value, list)
                else [str(value)]
            )

        cli_pairs = [
            cli_arguments(flags, argparse_ins)
            for flags, argparse_ins in self.get_arguments().items()
        ]
        return list(itertools.chain.from_iterable(cli_pairs))

    @abstractmethod
    def handle_arguments(self, args, **kwargs):
        """Post-process the parser's arguments

        Handle arguments from the given parser. The expectation is that the "args" namespace is taken in, processed, and
        a new namespace object is returned with the processed variants of the arguments.

        Args:
            args: arguments namespace of processed arguments
        Returns: namespace with processed results of arguments.
        """

    @classmethod
    def parse_known_args(
        cls,
        parser_classes,
        description="No tool description provided",
        arguments=None,
        **kwargs,
    ):
        """Parse and post-process arguments

        Create a parser for the given application using the description provided. This will then add all specified
        ParserBase subclasses' get_parser output as parent parses for the created parser. Then all of the handle
        arguments methods will be called, and the final namespace will be returned. This will allow unknown arguments
        which are returned as the last tuple result.

        Args:
            parser_classes: a list of ParserBase subclasses that will be used to
            description: description passed ot the argument parser
            arguments: arguments to process, None to use command line input
        Returns: namespace with all parsed arguments from all provided ParserBase subclasses
        """
        return cls._parse_args(
            parser_classes, description, arguments, use_parse_known=True, **kwargs
        )

    @classmethod
    def parse_args(
        cls,
        parser_classes,
        description="No tool description provided",
        arguments=None,
        **kwargs,
    ):
        """Parse and post-process arguments

        Create a parser for the given application using the description provided. This will then add all specified
        ParserBase subclasses' get_parser output as parent parses for the created parser. Then all of the handle
        arguments methods will be called, and the final namespace will be returned. This does not allow unknown
        arguments.

        Args:
            parser_classes: a list of ParserBase subclasses that will be used to
            description: description passed ot the argument parser
            arguments: arguments to process, None to use command line input
        Returns: namespace with all parsed arguments from all provided ParserBase subclasses
        """
        return cls._parse_args(parser_classes, description, arguments, **kwargs)

    @staticmethod
    def _parse_args(
        parser_classes,
        description="No tool description provided",
        arguments=None,
        use_parse_known=False,
        **kwargs,
    ):
        """Parse and post-process arguments helper

        Create a parser for the given application using the description provided. This will then add all specified
        ParserBase subclasses' get_parser output as parent parses for the created parser. Then all of the handle
        arguments methods will be called, and the final namespace will be returned.

        This takes a function that will take in a parser and return the parsing function to call on arguments.

        Args:
            parse_function_processor: takes a parser, returns the parse function to call
            parser_classes: a list of ParserBase subclasses that will be used to
            description: description passed ot the argument parser
            arguments: arguments to process, None to use command line input
            use_parse_known: use parse_known_arguments from argparse

        Returns: namespace with all parsed arguments from all provided ParserBase subclasses
        """
        composition = CompositeParser(parser_classes, description)
        parser = composition.get_parser()
        try:
            if use_parse_known:
                args_ns, *unknowns = parser.parse_known_args(arguments)
            else:
                args_ns = parser.parse_args(arguments)
                unknowns = []
            args_ns = composition.handle_arguments(args_ns, **kwargs)
        except ValueError as ver:
            print(f"[ERROR] Failed to parse arguments: {ver}", file=sys.stderr)
            parser.print_help()
            sys.exit(-1)
        except Exception as exc:
            print(f"[ERROR] {exc}", file=sys.stderr)
            sys.exit(-1)
        return args_ns, parser, *unknowns

    @staticmethod
    def find_in(token, deploy, is_file=True):
        """
        Find token in deploy directory by walking the directory looking for reg-ex. This effectively finds a file in a
        subtree and provides the path to it. Returns None when not found

        :param token: token to search for in the directory structure
        :param deploy: directory to start with
        :param is_file: true if looking for file, otherwise false
        :return: full path to token in tree
        """
        for dirpath, dirs, files in os.walk(deploy):
            for check in files if is_file else dirs:
                if re.match(f"^{str(token)}$", check):
                    return os.path.join(dirpath, check)
        return None


class ConfigDrivenParser(ParserBase):
    """Parser that allows options from configuration and command line

    This parser reads a configuration file (if supplied) and uses the values to drive the inputs to arguments. Command
    line arguments will still take precedence over the configured values.
    """

    DEFAULT_CONFIGURATION_PATH = Path("fprime-gds.yml")

    @classmethod
    def set_default_configuration(cls, path: Path):
        """Set path for (global) default configuration file

        Set the path for default configuration file. If unset, will use 'fprime-gds.yml'. Set to None to disable default
        configuration.
        """
        cls.DEFAULT_CONFIGURATION_PATH = path

    @classmethod
    def parse_args(
        cls,
        parser_classes,
        description="No tool description provided",
        arguments=None,
        **kwargs,
    ):
        """Parse and post-process arguments using inputs and config

        Parse the arguments in two stages: first parse the configuration data, ignoring unknown inputs, then parse the
        full argument set with the supplied configuration to fill in additional options.

        Args:
            parser_classes: a list of ParserBase subclasses that will be used to
            description: description passed ot the argument parser
            arguments: arguments to process, None to use command line input
        Returns: namespace with all parsed arguments from all provided ParserBase subclasses
        """
        arguments = sys.argv[1:] if arguments is None else arguments

        # Help should spill all the arguments, so delegate to the normal parsing flow including
        # this and supplied parsers
        if "-h" in arguments or "--help" in arguments:
            parsers = [ConfigDrivenParser] + parser_classes
            ParserBase.parse_args(parsers, description, arguments, **kwargs)
            sys.exit(0)

        # Custom flow involving parsing the arguments of this parser first, then passing the configured values
        # as part of the argument source
        ns_config, _, remaining = ParserBase.parse_known_args(
            [ConfigDrivenParser], description, arguments, **kwargs
        )
        config_options = ns_config.config_values.get("command-line-options", {})
        config_args = cls.flatten_options(config_options)
        # Argparse allows repeated (overridden) arguments, thus the CLI override is accomplished by providing
        # remaining arguments after the configured ones
        ns_full, parser = ParserBase.parse_args(
            parser_classes, description, config_args + remaining, **kwargs
        )
        ns_final = argparse.Namespace(**vars(ns_config), **vars(ns_full))
        return ns_final, parser

    @staticmethod
    def flatten_options(configured_options):
        """Flatten options down to arguments"""
        flattened = []
        for option, value in configured_options.items():
            flattened.append(f"--{option}")
            if value is not None:
                flattened.extend(
                    value if isinstance(value, (list, tuple)) else [f"{value}"]
                )
        return flattened

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Arguments needed for config processing"""
        return {
            ("-c", "--config"): {
                "dest": "config",
                "required": False,
                "default": self.DEFAULT_CONFIGURATION_PATH,
                "type": Path,
                "help": f"Argument configuration file path.",
            }
        }

    def handle_arguments(self, args, **kwargs):
        """Handle the arguments

        Loads the configuration file specified and fills in the `config_values` attribute of the namespace with the
        loaded configuration dictionary.
        """
        args.config_values = {}
        # Specified but non-existent config file is a hard error
        if (
            "-c" in sys.argv[1:] or "--config" in sys.argv[1:]
        ) and not args.config.exists():
            raise ValueError(
                f"Specified configuration file '{args.config}' does not exist"
            )
        # Read configuration if the file was set and exists
        if args.config is not None and args.config.exists():
            print(f"[INFO] Reading command-line configuration from: {args.config}")
            with open(args.config, "r") as file_handle:
                try:
                    loaded = yaml.safe_load(file_handle)
                    args.config_values = loaded if loaded is not None else {}
                except Exception as exc:
                    raise ValueError(
                        f"Malformed configuration {args.config}: {exc}", exc
                    )
        return args


class DetectionParser(ParserBase):
    """Parser that detects items from a root/directory or deployment"""

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Arguments needed for root processing"""
        return {
            ("-d", "--deployment"): {
                "dest": "deployment",
                "action": "store",
                "required": False,
                "type": str,
                "help": "Deployment installation/build output directory. [default: install_dest field in settings.ini]",
            }
        }

    def handle_arguments(self, args, **kwargs):
        """Handle the root, detecting it if necessary"""
        if args.deployment:
            args.deployment = Path(args.deployment)
            return args
        detected_toolchain = get_artifacts_root() / platform.system()
        if not detected_toolchain.exists():
            msg = f"{detected_toolchain} does not exist. Make sure to build."
            raise Exception(msg)
        likely_deployment = detected_toolchain / Path.cwd().name
        # Check if the deployment exists
        if likely_deployment.exists():
            args.deployment = likely_deployment
            return args
        child_directories = [
            child for child in detected_toolchain.iterdir() if child.is_dir()
        ]
        if not child_directories:
            msg = f"No deployments found in {detected_toolchain}. Specify deployment with: --deployment"
            raise Exception(msg)
        # Works for the old structure where the bin, lib, and dict directories live immediately under the platform
        elif len(child_directories) == 3 and set(
            [path.name for path in child_directories]
        ) == {"bin", "lib", "dict"}:
            args.deployment = detected_toolchain
            return args
        elif len(child_directories) > 1:
            msg = f"Multiple deployments found in {detected_toolchain}. Choose using: --deployment"
            raise Exception(msg)
        args.deployment = child_directories[0]
        return args


class BareArgumentParser(ParserBase):
    """Takes in the argument specification (used in plugins and get_arguments) to parse args

    This parser takes in and uses a raw specification of arguments as seen in plugins and arguments to perform argument
    parsing. The spec is a map of flag tuples to argparse kwargs.

    Argument handling only validates using the checking_function which is a function taking in keyword arguments for
    each cli argument specified. This function will be called as such: `checking_function(**args)`. Use None to skip
    argument checking.  checking_function should raise ValueError to indicate an error with an argument.
    """

    def __init__(self, specification, checking_function=None):
        """Initialize this parser with the provided specification"""
        self.specification = specification
        self.checking_function = checking_function

    def get_arguments(self):
        """Raw specification is returned immediately"""
        return self.specification

    def handle_arguments(self, args, **kwargs):
        """Handle argument calls checking function to validate"""
        if self.checking_function is not None:
            self.checking_function(**self.extract_arguments(args))
        return args

    def extract_arguments(self, args) -> Dict[str, Any]:
        """Extract argument values from the args namespace into a map matching the original specification

        This function extracts arguments matching the original specification and returns them as a dictionary of key-
        value pairs.

        Return:
            filled arguments dictionary
        """
        expected_args = self.specification
        argument_destinations = [
            (
                value["dest"]
                if "dest" in value
                else key[0].replace("--", "").replace("-", "_")
            )
            for key, value in expected_args.items()
        ]
        filled_arguments = {
            destination: getattr(args, destination)
            for destination in argument_destinations
        }
        return filled_arguments


class IndividualPluginParser(BareArgumentParser):
    """Parser for an individual plugin's command line

    A CLI parser for an individual plugin. This handles all the functions and arguments that apply to the parsing of a
    single plugin's arguments. It also handles FEATURE plugin disable flags.
    """

    def __init__(self, plugin_system: Plugins, plugin_class: type):
        """Initialize the plugin parser

        Args:
            plugin_system: Plugins object used to work with the plugin system
            plugin_class: plugin class used for this specific parser
        """
        # Add disable flags for feature type plugins
        super().__init__(plugin_class.get_arguments(), plugin_class.check_arguments)
        self.disable_flag_destination = (
            f"disable-{plugin_class.get_name()}".lower().replace("-", "_")
        )
        self.plugin_class = plugin_class
        self.plugin_system = plugin_system

    def get_arguments(self):
        """Get the arguments for this plugin

        The individual plugin parser will read the arguments from the supplied plugin class. Additionally, if the
        plugin_class's plugin_type is FEATURE then this parser will add an disable flag to allow users to turn disable
        the plugin feature.
        """
        arguments = {}
        if self.plugin_class.type == PluginType.FEATURE:
            arguments.update(
                {
                    (f"--disable-{self.plugin_class.get_name()}",): {
                        "action": "store_true",
                        "default": False,
                        "dest": self.disable_flag_destination,
                        "help": f"Disable the {self.plugin_class.category} plugin '{self.plugin_class.get_name()}'",
                    }
                }
            )
        arguments.update(super().get_arguments())
        return arguments

    def handle_arguments(self, args, **kwargs):
        """Handle the given arguments for a plugin

        This will process the arguments for a given plugin. Additionally, it will construct the plugin object and
        supply the constructed object to the plugin system if the plugin is a selection or is enabled.

        Args:
            args: argparse namespace
        """
        arguments = super().handle_arguments(
            args, **kwargs
        )  # Perform argument checking first
        if not getattr(args, self.disable_flag_destination, False):
            # Remove the disable flag from the arguments
            plugin_arguments = {
                key: value
                for key, value in self.extract_arguments(arguments).items()
                if key != self.disable_flag_destination
            }

            plugin_zero_argument_class = functools.partial(
                self.plugin_class.get_implementor(), **plugin_arguments
            )
            self.plugin_system.add_bound_class(
                self.plugin_class.category, plugin_zero_argument_class
            )
        return arguments

    def get_plugin_class(self):
        """Plugin class accessor"""
        return self.plugin_class


class PluginArgumentParser(ParserBase):
    """Parser for arguments coming from plugins"""

    DESCRIPTION = "Plugin options"
    # Defaults:
    FPRIME_CHOICES = {
        "framing": "space-packet-space-data-link",
        "communication": "ip",
    }

    def __init__(self, plugin_system: Plugins = None):
        """Initialize the plugin information for this parser

        This will initialize this plugin argument parser with the supplied plugin system. If not supplied this will use
        the system plugin singleton, which is configured elsewhere.
        """
        # Accept the supplied plugin system defaulting to the global singleton
        self.plugin_system = plugin_system if plugin_system else Plugins.system()
        self._plugin_map = {
            category: [
                IndividualPluginParser(self.plugin_system, plugin)
                for plugin in self.plugin_system.get_plugins(category)
            ]
            for category in self.plugin_system.get_categories()
        }

    def fill_parser(self, parser):
        """Fill supplied parser with grouped arguments

        Fill the supplied parser with arguments from the `get_arguments` method invocation. This implementation groups
        arguments based on the constituent parser that the argument comes from. Category specific arguments are also
        added (i.e. SELECTION type selection arguments).

        Args:
            parser: parser to fill
        """
        for category, plugin_parsers in self._plugin_map.items():
            # Add category specific flags (selection flags, etc)
            argument_group = parser.add_argument_group(
                title=f"{category.title()} Plugin Options"
            )
            self.add_arguments_from_specification(
                argument_group, self.get_category_arguments(category)
            )

            # Handle the individual plugin parsers
            for plugin_parser in plugin_parsers:
                plugin = plugin_parser.get_plugin_class()
                argument_group = parser.add_argument_group(
                    title=f"{category.title()} Plugin '{plugin.get_name()}' Options"
                )
                plugin_parser.fill_parser(argument_group)

    def get_category_arguments(self, category):
        """Get category arguments for a given plugin category

        This function will generate category arguments for the supplied category. These arguments will follow the
        standard argument specification of a dictionary of flag tuples to argparse keyword arguments.

        Currently category specific arguments are just selection flags for SELECTION type plugins.

        Args:
            category: category arguments
        """
        plugin_type = self.plugin_system.get_category_plugin_type(category)
        plugins = [
            plugin_parser.get_plugin_class()
            for plugin_parser in self._plugin_map[category]
        ]

        arguments: Dict[Tuple[str, ...], Dict[str, Any]] = {}

        # Add category options: SELECTION plugins add a selection flag
        if plugin_type == PluginType.SELECTION:
            arguments.update(
                {
                    (f"--{category}-selection",): {
                        "choices": [choice.get_name() for choice in plugins],
                        "help": f"Select {category} implementer.",
                        "default": self.FPRIME_CHOICES.get(
                            category, list(plugins)[0].get_name()
                        ),
                    }
                }
            )
        return arguments

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments to used in plugin system

        This will return the command line arguments all the plugins contained within the supplied plugin system. This
        will recursively return plugins from all of the IndividualPluginParser objects composing this plugin argument
        parser. Arguments are returned in the standard specification form of tuple of flags mapped to a dictionary of
        argparse kwarg inputs.
        """
        arguments: Dict[Tuple[str, ...], Dict[str, Any]] = {}
        for category, plugin_parsers in self._plugin_map.items():
            arguments.update(self.get_category_arguments(category))
            [
                arguments.update(plugin_parser.get_arguments())
                for plugin_parser in plugin_parsers
            ]
        return arguments

    def handle_arguments(self, args, **kwargs):
        """Handle the plugin arguments

        This will handle the plugin arguments delegating each to the IndividualPluginParser. For SELECTION plugins this
        will bind a single instance of the selected plugin to its arguments. For FEATURE plugins it will bind arguments
        to every enabled plugin. Bound plugins are registered with the plugin system.
        """
        for category, plugin_parsers in self._plugin_map.items():
            plugin_type = self.plugin_system.get_category_plugin_type(category)
            self.plugin_system.start_loading(category)
            # Selection plugins choose one plugin and instantiate it
            if plugin_type == PluginType.SELECTION:
                try:
                    self.plugin_system.get_selected_class(category)
                except PluginsNotLoadedException:
                    selection_string = getattr(args, f"{category}_selection")
                    matching_plugin_parsers = [
                        plugin_parser
                        for plugin_parser in plugin_parsers
                        if plugin_parser.get_plugin_class().get_name()
                        == selection_string
                    ]
                    assert (
                        len(matching_plugin_parsers) == 1
                    ), "Plugin selection system failed"
                    args = matching_plugin_parsers[0].handle_arguments(args, **kwargs)
            # Feature plugins instantiate all enabled plugins
            elif plugin_type == PluginType.FEATURE:
                for plugin_parser in plugin_parsers:
                    args = plugin_parser.handle_arguments(args, **kwargs)
        return args


class CompositeParser(ParserBase):
    """Composite parser handles parsing as a composition of multiple other parsers"""

    def __init__(self, constituents, description=None):
        """Construct this parser by instantiating the sub-parsers"""
        self.given = description
        constructed = [
            constituent() if callable(constituent) else constituent
            for constituent in constituents
        ]
        # Check to ensure everything passed in became a ParserBase after construction
        for i, construct in enumerate(constructed):
            assert isinstance(
                construct, ParserBase
            ), f"{construct.__class__.__name__} ({i}) not a ParserBase child"
        flattened = [
            item.constituents if isinstance(item, CompositeParser) else [item]
            for item in constructed
        ]
        self.constituent_parsers = {*itertools.chain.from_iterable(flattened)}

    def fill_parser(self, parser):
        """File supplied parser with grouped arguments

        Fill the supplied parser with arguments from the `get_arguments` method invocation. This implementation groups
        arguments based on the constituent that sources the argument.

        Args:
            parser: parser to fill
        """
        for constituent in sorted(self.constituents, key=lambda x: x.description):
            if isinstance(constituent, (PluginArgumentParser, CompositeParser)):
                constituent.fill_parser(parser)
            else:
                argument_group = parser.add_argument_group(
                    title=constituent.description
                )
                constituent.fill_parser(argument_group)

    @property
    def constituents(self):
        """Get constituent"""
        return self.constituent_parsers

    @property
    def description(self):
        """Return parser description"""
        return (
            self.given
            if self.given
            else ",".join(item.description for item in self.constituents)
        )

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Get the argument from all constituents"""
        arguments = {}
        for constituent in self.constituents:
            arguments.update(constituent.get_arguments())
        return arguments

    def handle_arguments(self, args, **kwargs):
        """Process all constituent arguments"""
        for constituent in self.constituents:
            args = constituent.handle_arguments(args, **kwargs)
        return args


class CommExtraParser(ParserBase):
    """Parses extra communication arguments"""

    DESCRIPTION = "Communications options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Get arguments for the comm-layer parser"""
        com_arguments = {
            ("--output-unframed-data",): {
                "dest": "output_unframed_data",
                "action": "store",
                "nargs": "?",
                "help": "Log unframed data to supplied file relative to log directory. Use '-' for standard out.",
                "default": None,
                "const": "unframed.log",
                "required": False,
            },
        }
        return com_arguments

    def handle_arguments(self, args, **kwargs):
        return args


class LogDeployParser(ParserBase):
    """
    A parser that handles log files by reading in a '--logs' directory or a '--deploy' directory to put the logs into
    as a default. This is useful as a parsing fragment for any application that produces log files and needs these logs
    to end up in the proper place.
    """

    DESCRIPTION = "Logging options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments to parse logging options"""
        return {
            ("-l", "--logs"): {
                "dest": "logs",
                "action": "store",
                "default": os.path.join(os.getcwd(), "logs"),
                "type": str,
                "help": "Logging directory. Created if non-existent. [default: %(default)s]",
            },
            ("--log-directly",): {
                "dest": "log_directly",
                "action": "store_true",
                "default": False,
                "help": "Logging directory is used directly, no extra dated directories created.",
            },
            ("--log-to-stdout",): {
                "action": "store_true",
                "default": False,
                "help": "Log to standard out along with log output files",
            },
            ("--log-level-gds",): {
                "action": "store",
                "dest": "log_level_gds",
                "choices": ["DEBUG", "INFO", "WARNING", "ERROR"],
                "default": "INFO",
                "help": "Set the logging level of GDS processes [default: %(default)s]",
            },
            ("--disable-data-logging",): {
                "action": "store_true",
                "default": False,
                "help": "Disable logging of each data item",
            },
        }

    def handle_arguments(self, args, **kwargs):
        """
        Read the arguments specified in this parser and validate the expected inputs.

        :param args: parsed arguments as namespace
        :return: args namespace
        """
        # Get logging dir
        if not args.log_directly:
            args.logs = os.path.abspath(
                os.path.join(
                    args.logs, datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
                )
            )
            # A dated directory has been set, all log handling must now be direct
            args.log_directly = True

        # Make sure directory exists
        try:
            os.makedirs(args.logs, exist_ok=True)
        except OSError as osexc:
            if osexc.errno != errno.EEXIST:
                raise
        # Setup the basic python logging
        fprime_gds.common.logger.configure_py_log(
            args.logs, mirror_to_stdout=args.log_to_stdout, log_level=args.log_level_gds
        )
        return args


class MiddleWareParser(ParserBase):
    """
    Middleware (ThreadedTcpServer, ZMQ) interface that looks for an address and a port. The argument handling will
    attempt to connect to the socket to ensure that it is a valid address/port and report any errors. This is then
    immediately closes the port after use. There is a minor race-condition between this check and the actual usage,
    however; it should be close enough.
    """

    DESCRIPTION = "Middleware options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to run a and connect to the GDS middleware"""
        zmq_arguments = {
            ("--no-zmq",): {
                "dest": "zmq",
                "action": "store_false",
                "help": "Disable ZMQ transportation layer, falling back to TCP socket server.",
                "default": True,
            },
            ("--zmq-transport",): {
                "dest": "zmq_transport",
                "nargs": 2,
                "help": "Pair of URls used with --zmq to setup ZeroMQ transportation [default: %(default)s]",
                "default": [
                    "ipc:///tmp/fprime-server-in",
                    "ipc:///tmp/fprime-server-out",
                ],
                "metavar": ("serverInUrl", "serverOutUrl"),
            },
        }
        tts_arguments = {
            ("--tts-port",): {
                "dest": "tts_port",
                "action": "store",
                "type": int,
                "help": "Set the threaded TCP socket server port when ZMQ is not used [default: %(default)s]",
                "default": 50050,
            },
            ("--tts-addr",): {
                "dest": "tts_addr",
                "action": "store",
                "type": str,
                "help": "Set the threaded TCP socket server address when ZMQ is not used [default: %(default)s]",
                "default": "0.0.0.0",
            },
        }
        return {**zmq_arguments, **tts_arguments}

    def handle_arguments(self, args, **kwargs):
        """
        Checks to ensure that the specified port and address is available before connecting. This prevents user from
        attempting to run on a port that is unavailable.

        :param args: parsed argument namespace
        :return: args namespace
        """
        is_client = kwargs.get("client", False)
        tts_connection_address = (
            args.tts_addr.replace("0.0.0.0", "127.0.0.1")
            if is_client
            else args.tts_addr
        )

        args.connection_uri = f"tcp://{tts_connection_address}:{args.tts_port}"
        args.connection_transport = ThreadedTCPSocketClient
        if args.zmq:
            args.connection_uri = args.zmq_transport
            args.connection_transport = ZmqClient
        elif not is_client:
            check_port(args.tts_addr, args.tts_port)
        return args


class DictionaryParser(DetectionParser):
    """Parser for deployments"""

    DESCRIPTION = "Dictionary options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Arguments to handle deployments"""
        return {
            **super().get_arguments(),
            **{
                ("--dictionary",): {
                    "dest": "dictionary",
                    "action": "store",
                    "default": None,
                    "required": False,
                    "type": str,
                    "help": "Path to dictionary. Overrides automatic dictionary detection.",
                },
                ("--packet-spec",): {
                    "dest": "packet_spec",
                    "action": "store",
                    "default": None,
                    "required": False,
                    "type": str,
                    "help": "Path to packet XML specification (should not be used if JSON packet definitions are used).",
                },
                ("--packet-set-name",): {
                    "dest": "packet_set_name",
                    "action": "store",
                    "default": None,
                    "required": False,
                    "type": str,
                    "help": "Name of packet set defined in the JSON dictionary.",
                },
            },
        }

    def handle_arguments(self, args, **kwargs):
        """Handle arguments as parsed"""
        # Find dictionary setting via "dictionary" argument or the "deploy" argument
        if args.dictionary is not None and not os.path.exists(args.dictionary):
            msg = f"Dictionary file {args.dictionary} does not exist"
            raise ValueError(msg)
        elif args.dictionary is None:
            args = super().handle_arguments(args, **kwargs)
            args.dictionary = find_dict(args.deployment)
        return args


class FileHandlingParser(ParserBase):
    """Parser for deployments"""

    DESCRIPTION = "File handling options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Arguments to handle deployments"""

        username = getpass.getuser()

        return {
            ("--file-storage-directory",): {
                "dest": "files_storage_directory",
                "action": "store",
                "default": "/tmp/" + username,
                "required": False,
                "type": str,
                "help": "Directory to store uplink and downlink files. Default: %(default)s",
            },
            ("--remote-sequence-directory",): {
                "dest": "remote_sequence_directory",
                "action": "store",
                "default": "/seq",
                "required": False,
                "type": str,
                "help": "Directory to save command sequence binaries, on the remote FSW. Default: %(default)s",
            },
        }

    def handle_arguments(self, args, **kwargs):
        """Handle arguments as parsed"""
        try:
            Path(args.files_storage_directory).mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise PermissionError(
                f"{args.files_storage_directory} is not writable. Fix permissions or change storage directory with --file-storage-directory."
            )
        return args


class StandardPipelineParser(CompositeParser):
    """Standard pipeline argument parser: combination of MiddleWare and"""

    CONSTITUENTS = [
        DictionaryParser,
        FileHandlingParser,
        MiddleWareParser,
        LogDeployParser,
    ]

    def __init__(self):
        """Initialization"""
        super().__init__(
            constituents=self.CONSTITUENTS, description="Standard pipeline setup"
        )

    @staticmethod
    def pipeline_factory(args_ns, pipeline=None) -> StandardPipeline:
        """A factory of the standard pipeline given the handled arguments"""
        pipeline_arguments = {
            "config": ConfigManager(),
            "dictionary": args_ns.dictionary,
            "file_store": args_ns.files_storage_directory,
            "packet_spec": args_ns.packet_spec,
            "packet_set_name": args_ns.packet_set_name,
            "logging_prefix": args_ns.logs,
            "data_logging_enabled": not args_ns.disable_data_logging
        }
        pipeline = pipeline if pipeline else StandardPipeline()
        pipeline.transport_implementation = args_ns.connection_transport
        try:
            pipeline.setup(**pipeline_arguments)
            pipeline.connect(args_ns.connection_uri)
        except Exception:
            # In all error cases, pipeline should be shutdown before continuing with exception handling
            try:
                pipeline.disconnect()
            finally:
                raise
        return pipeline


class CommParser(CompositeParser):
    """Comm Executable Parser"""

    CONSTITUENTS = [
        CommExtraParser,
        MiddleWareParser,
        LogDeployParser,
    ]

    def __init__(self):
        """Initialization"""
        # Added here to ensure the call to Plugins does not interfere with the full plugin system
        comm_plugin_parser_instance = PluginArgumentParser(
            Plugins(["communication", "framing"])
        )
        super().__init__(
            constituents=self.CONSTITUENTS + [comm_plugin_parser_instance],
            description="Communications bridge application",
        )


class GdsParser(ParserBase):
    """
    Provides a parser for the following arguments:

    - dictionary: path to dictionary, either a folder for py_dicts, or a file for XML dicts
    - logs: path to logging path
    - config: configuration for GDS.

    Note: deployment can help in setting both dictionary and logs, but isn't strictly required.
    """

    DESCRIPTION = "GUI options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to run a binary deployment via the GDS"""
        return {
            ("-g", "--gui"): {
                "choices": GUIS,
                "dest": "gui",
                "type": str,
                "help": "Set the desired GUI system for running the deployment. [default: %(default)s]",
                "default": "html",
            },
            ("--gui-addr",): {
                "dest": "gui_addr",
                "action": "store",
                "default": "127.0.0.1",
                "required": False,
                "type": str,
                "help": "Set the GUI server address [default: %(default)s]",
            },
            ("--gui-port",): {
                "dest": "gui_port",
                "action": "store",
                "default": "5000",
                "required": False,
                "type": str,
                "help": "Set the GUI server address [default: %(default)s]",
            },
        }

    def handle_arguments(self, args, **kwargs):
        """
        Takes the arguments from the parser, and processes them into the needed map of key to dictionaries for the
        program. This will throw if there is an error.

        :param args: parsed args into a namespace
        :return: args namespace
        """
        return args


class BinaryDeployment(DetectionParser):
    """
    Parsing subclass used to read the arguments of the binary application. This derives functionality from a comm parser
    and represents the flight-side of the equation.
    """

    DESCRIPTION = "FPrime binary options"

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to run a binary deployment via the GDS"""
        return {
            **super().get_arguments(),
            **{
                ("-n", "--no-app"): {
                    "dest": "noapp",
                    "action": "store_true",
                    "default": False,
                    "help": "Do not run deployment binary. Overrides --app.",
                },
                ("--app",): {
                    "dest": "app",
                    "action": "store",
                    "required": False,
                    "type": str,
                    "help": "Path to app to run. Overrides automatic app detection.",
                },
            },
        }

    def handle_arguments(self, args, **kwargs):
        """
        Takes the arguments from the parser, and processes them into the needed map of key to dictionaries for the
        program. This will throw if there is an error.

        :param args: parsed arguments in namespace
        :return: args namespaces
        """
        # No app, stop processing now
        if args.noapp:
            return args
        args = super().handle_arguments(args, **kwargs)
        args.app = Path(args.app) if args.app else Path(find_app(args.deployment))
        if not args.app.is_file():
            msg = f"F prime binary '{args.app}' does not exist or is not a file"
            raise ValueError(msg)
        return args


class SearchArgumentsParser(ParserBase):
    """Parser for search arguments"""

    DESCRIPTION = "Searching and filtering options"

    def __init__(self, command_name: str) -> None:
        self.command_name = command_name

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments necessary to search through channels/events/commands"""
        return {
            ("--list",): {
                "dest": "is_printing_list",
                "action": "store_true",
                "help": f"list all possible {self.command_name[:-1]} types the current F Prime instance could produce, based on the {self.command_name} dictionary, sorted by {self.command_name[:-1]} type ID",
            },
            ("-i", "--ids"): {
                "dest": "ids",
                "action": "store",
                "required": False,
                "type": int,
                "nargs": "+",
                "help": f"only show {self.command_name} matching the given type ID(s) 'ID'; can provide multiple IDs to show all given types",
                "metavar": "ID",
            },
            ("-c", "--components"): {
                "dest": "components",
                "nargs": "+",
                "required": False,
                "type": str,
                "help": f"only show {self.command_name} from the given component name 'COMP'; can provide multiple components to show {self.command_name} from all components given",
                "metavar": "COMP",
            },
            ("-s", "--search"): {
                "dest": "search",
                "required": False,
                "type": str,
                "help": f'only show {self.command_name} whose name or output string exactly matches or contains the entire given string "STRING"',
            },
        }

    def handle_arguments(self, args, **kwargs):
        return args


class RetrievalArgumentsParser(ParserBase):
    """Parser for retrieval arguments"""

    DESCRIPTION = "Data retrieval options"

    def __init__(self, command_name: str) -> None:
        self.command_name = command_name

    def get_arguments(self) -> Dict[Tuple[str, ...], Dict[str, Any]]:
        """Return arguments to retrieve channels/events/commands in specific ways"""
        return {
            ("-t", "--timeout"): {
                "dest": "timeout",
                "action": "store",
                "required": False,
                "type": float,
                "help": f"wait at most SECONDS seconds for a single new {self.command_name[:-1]}, then exit (defaults to listening until the user exits via CTRL+C, and logging all {self.command_name})",
                "metavar": "SECONDS",
                "default": 0.0,
            },
            ("-j", "--json"): {
                "dest": "json",
                "action": "store_true",
                "required": False,
                "help": "returns response in JSON format",
            },
        }

    def handle_arguments(self, args, **kwargs):
        return args



"""
@brief file writer to write decoded objects into files

This file writer takes in decoded data and writes the data
to the correct log files and destination files

@date Created August 8, 2019
@author Blake A. Harriman

@bug No known bugs
"""

import logging
import os

import fprime.constants

import fprime_gds.common.handlers
from fprime_gds.common.data_types.file_data import FilePacketType
from fprime_gds.common.files.helpers import (
    FileStates,
    TransmitFile,
    TransmitFileState,
    file_to_dict,
)

LOGGER = logging.getLogger("downlink")
LOGGER.setLevel(logging.INFO)


class FileDownlinker(fprime_gds.common.handlers.DataHandler):
    """File writer class for decoded packets"""

    def __init__(self, directory, timeout=20.0, log_dir=None):
        """
        FileWriter class constructor

        Args:
            directory: directory in which to write downlinked files
            timeout: This is used to keep track of how long the program should wait before throwing
            a timeout exception:

        Returns:
            An initialized FileWriter object.
        """
        super().__init__()
        self.__directory = directory
        self.__log_dir = log_dir if log_dir is not None else directory
        self.active = None
        self.files = []
        self.state = FileStates.IDLE
        self.sequence = 0  # Keeps track of what the current sequence ID should be
        self.timer = fprime_gds.common.files.helpers.Timeout()
        self.timer.setup(self.timeout, timeout)

    def data_callback(self, data, sender=None):
        """
        Function called to pass data to the writer class

        Args:
            data: Binary data that has been decoded and passed to the correct consumer
        """
        self.timer.restart()  # Start or restart the timer
        packet_type = data.packetType
        # Check the packet type, and route to the appropriate sub-function
        if packet_type == FilePacketType.START:
            self.handle_start(data)
        elif packet_type == FilePacketType.DATA:
            self.handle_data(data)
        elif packet_type == FilePacketType.END:
            self.handle_end(data)
        elif packet_type == FilePacketType.CANCEL:
            self.handle_cancel(data)
        else:
            LOGGER.warning("Invalid file detected descriptor detected: %d", packet_type)

    def handle_start(self, data):
        """
        Handle a start packet data type.

        :param data: data packet that is a start packet
        """
        # Initialize all relevant START packet attributes into variables from file_data
        size = data.size
        source_path = data.sourcePath.decode(fprime.constants.DATA_ENCODING)
        dest_path = data.destPath.decode(fprime.constants.DATA_ENCODING)
        if self.state != FileStates.IDLE:
            LOGGER.warning("File transfer already inprogress. Aborting original.")
            self.finish()
        # Create the destination file where the DATA packet data will be stored
        assert self.active is None, "File is already open, something went wrong"
        self.active = TransmitFile(
            source_path,
            os.path.join(self.__directory, self.sanitize(dest_path)),
            size,
            self.__log_dir,
        )
        try:
            self.active.open(TransmitFileState.WRITE)
        except PermissionError as exc:
            self.state = FileStates.ERROR
            LOGGER.warning(
                "Unable to open file for writing. Discarding subsequent downlink packets. "
                + str(exc)
            )
            return
        LOGGER.addHandler(self.active.log_handler)
        message = "Received START packet with metadata:\n"
        message += "\tSize: %d\n"
        message += "\tSource: %s\n"
        message += "\tDestination: %s"
        LOGGER.info(message, size, source_path, dest_path)
        self.files.append(self.active)
        self.state = FileStates.RUNNING
        self.sequence += 1

    def handle_data(self, data):
        """
        Handle the data packet.

        :param data: data packet
        """
        # Initialize all relevant DATA packet attributes into variables from file_data
        offset = data.offset
        if self.state != FileStates.RUNNING:
            # ERROR state means GDS is not ready to receive data packets (permission error)
            # No need to log this, as it is already logged in handle_start and would only spam logs
            if self.state != FileStates.ERROR:
                LOGGER.warning("Received unexpected data packet for offset: %d", offset)
        else:
            if data.seqID != self.sequence:
                LOGGER.warning(
                    "Data packet has unexpected sequence id. Expected: %d got %d",
                    self.sequence,
                    data.seqID,
                )
            data_bytes = data.dataVar
            # Write the data information to the file
            self.active.write(data_bytes, offset)
            self.active.seek = offset + len(data_bytes)
            LOGGER.info(
                "Received DATA packet writing %d bytes to offset %s",
                len(data_bytes),
                offset,
            )
        self.sequence += 1

    def handle_cancel(self, _):
        """
        Handle cancel packet.

        :param _: data unused in cancel packet
        """
        # CANCEL Packets have no data
        if self.state != FileStates.RUNNING:
            LOGGER.warning("Received unexpected cancel packet")
        else:
            LOGGER.info("Received CANCEL packet, stopping downlink")
            self.finish()

    def handle_end(self, data):
        """
        Handle the end packet.

        :param data: end packet
        """
        # Initialize all relevant END packet attributes into variables from file_data
        # hashValue attribute is not relevant right now, but might be in the future
        if self.state == FileStates.ERROR:
            LOGGER.info("Received END packet for discarded downlink")
            self.finish()
        elif self.state == FileStates.RUNNING:
            if data.seqID != self.sequence:
                LOGGER.warning(
                    "End packet has unexpected sequence id. Expected: %d got %d",
                    self.sequence,
                    data.seqID,
                )
            LOGGER.info("Received END packet, finishing downlink")
            self.finish()
        else:
            LOGGER.warning("Received unexpected END packet")

    def timeout(self):
        """Timeout the current downlink"""
        LOGGER.warning("Timeout while downlinking file, aborting")
        self.finish()

    def finish(self):
        """Finish the file downlink"""
        self.timer.stop()
        self.state = FileStates.IDLE
        self.sequence = 0
        self.active.state = "FINISHED"
        LOGGER.removeHandler(self.active.log_handler)
        self.active.close()
        self.active = None

    def current_files(self):
        """Return the current list of downlinked files"""
        return file_to_dict(self.files, uplink=False)

    @staticmethod
    def sanitize(filename):
        """
        Sanitize the given filename by removing slashes that would make new directories.

        :param filename: filename to sanitize
        :return: sanitized filename
        """
        return filename.replace(os.sep, "_")

    @property
    def directory(self):
        return self.__directory



""" ccsds.apid: APID mapping functions for F´ data """
from fprime_gds.common.utils.data_desc_type import DataDescType
from fprime.common.models.serialize.numerical_types import U32Type

class APID(object):
    """ APID implementations """
    #TODO: use the DataDescType configured by loading the dictionary

    @classmethod
    def from_type(cls, data_type: DataDescType):
        """ Map from data description type to APID """
        return data_type.value

    @classmethod
    def from_data(cls, data):
        """ Map from data bytes to APID """
        u32_type = U32Type()
        u32_type.deserialize(data, offset=0)
        return cls.from_type(DataDescType(u32_type.val))



"""
DataDescType:

Defines an enumeration that represents each type of data packet that can be downlinked.
"""
from enum import Enum


# TODO: these values should be read from the dictionary instead of hardcoded here

DataDescType = Enum(
    "DataDescType",
    {
        # Command packet type - incoming
        "FW_PACKET_COMMAND": 0,
        # Telemetry packet type - outgoing
        "FW_PACKET_TELEM": 1,
        # Log type - outgoing
        "FW_PACKET_LOG": 2,
        # File type - incoming and outgoing
        "FW_PACKET_FILE": 3,
        # Packetized telemetry packet type
        "FW_PACKET_PACKETIZED_TLM": 4,
        # Idle packet
        "FW_PACKET_IDLE": 5,
        # Handshake packet
        "FW_PACKET_HAND": 0xFE,
        # Unknown packet
        "FW_PACKET_UNKNOWN": 0xFF,
        # Space Packet Idle APID
        "CCSDS_SPACE_PACKET_IDLE_APID": 0x7FF,
    },
)



#!/usr/bin/env python3
# ===============================================================================
# NAME: tinyseqgen
#
# DESCRIPTION: A tiny sequence generator for F Prime. This sequence compiler takes a
# .seq file as input and produces a binary sequence file compatible with the
# F Prime sequence file loader and sequence file runner.
# AUTHOR: Kevin Dinkel
# EMAIL:  dinkel@jpl.nasa.gov
# DATE CREATED: December 15, 2015
#
# Copyright 2015, California Institute of Technology.
# ALL RIGHTS RESERVED. U.S. Government Sponsorship acknowledged.
# ===============================================================================

import argparse
import os
import sys

from fprime.common.models.serialize.time_type import TimeBase, TimeType

from fprime_gds.common.data_types import exceptions as gseExceptions
from fprime_gds.common.data_types.cmd_data import CmdData, CommandArgumentsException
from fprime_gds.common.encoders.seq_writer import SeqBinaryWriter
from fprime_gds.common.loaders.cmd_json_loader import CmdJsonLoader
from fprime_gds.common.parsers.seq_file_parser import SeqFileParser

__author__ = "Tim Canham"
__version__ = "1.0"
__email__ = "timothy.canham@jpl.nasa.gov"


class SeqGenException(gseExceptions.GseControllerException):
    def __init__(self, val):
        super().__init__(str(val))


# except:
#  __error("The Gse source code was not found in your $PYTHONPATH variable. Please set PYTHONPATH to something like: $BUILD_ROOT/Gse/src:$BUILD_ROOT/Gse/generated/$DEPLOYMENT_NAME")


def generateSequence(inputFile, outputFile, dictionary, timebase, cont=False):
    """
    Write a binary sequence file from a text sequence file
    @param inputFile: A text input sequence file name (usually a .seq extension)
    @param outputFile: An output binary sequence file name (usually a .bin extension)
    """

    # Check for files
    if not os.path.isfile(inputFile):
        msg = f"Can't open file '{inputFile}'. "
        raise SeqGenException(msg)

    if not os.path.isfile(dictionary):
        msg = f"Can't open file '{dictionary}'. "
        raise SeqGenException(msg)

    # Check the user environment:
    cmd_json_dict = CmdJsonLoader(dictionary)
    try:
        (cmd_id_dict, cmd_name_dict, versions) = cmd_json_dict.construct_dicts(
            dictionary
        )
    except gseExceptions.GseControllerUndefinedFileException:
        msg = f"Can't open file '{dictionary}'. "
        raise SeqGenException(msg)

    # Parse the input file:
    command_list = []
    file_parser = SeqFileParser()

    parsed_seq = file_parser.parse(inputFile, cont=cont)

    messages = []
    try:
        for i, descriptor, seconds, useconds, mnemonic, args in parsed_seq:
            try:
                if mnemonic not in cmd_name_dict:
                    msg = f"Line {i + 1}: '{mnemonic}' does not match any command in the command dictionary."
                    raise SeqGenException(
                        msg
                    )
                # Set the command arguments:
                try:
                    cmd_time = TimeType(
                        TimeBase["TB_DONT_CARE"].value,
                        seconds=seconds,
                        useconds=useconds,
                    )
                    cmd_data = CmdData(
                        args,
                        cmd_name_dict[mnemonic],
                        cmd_desc=descriptor,
                        cmd_time=cmd_time,
                    )
                except CommandArgumentsException as e:
                    msg = f"Line {i + 1}: {mnemonic} errored: {','.join(e.errors)}"
                    raise SeqGenException(
                        msg
                    )
                command_list.append(cmd_data)
            except SeqGenException as exc:
                if not cont:
                    raise
                messages.append(exc.getMsg())
    except gseExceptions.GseControllerParsingException as e:
        raise SeqGenException("\n".join([e.getMsg()] + messages))
    if cont and messages:
        raise SeqGenException("\n".join(messages))

    # Write to the output file:
    writer = SeqBinaryWriter(timebase=timebase)
    if not outputFile:
        outputFile = f"{os.path.splitext(inputFile)[0]}.bin"
    try:
        writer.open(outputFile)
    except:
        msg = f"Encountered problem opening output file '{outputFile}'."
        raise SeqGenException(
            msg
        )

    writer.write(command_list)
    writer.close()


help_text = "seqgen.py -d"


def main():
    """
    The main program if run from the command line. Note that this file can also be used
    as a module by calling the generateSequence() function
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "sequence", action="store", type=str, help="Path to input sequence file"
    )
    parser.add_argument(
        "output",
        action="store",
        nargs="?",
        type=str,
        help="Path to output binary file",
        default=None,
    )

    parser.add_argument(
        "-d",
        "--dictionary",
        dest="dictionary",
        action="store",
        type=str,
        required=True,
        help="JSON Dictionary file name",
    )
    parser.add_argument(
        "-t",
        "--timebase",
        dest="timebase",
        action="store",
        type=str,
        default=None,
        help="Set base path to generated command/telemetry definition files [default: any]",
    )

    opts = parser.parse_args()

    if opts.timebase is None:
        timebase = 0xFFFF
    else:
        try:
            timebase = int(opts.timebase, 0)
        except ValueError:
            print(f"Could not parse time base {opts.timebase}")
            return 1

    inputfile = opts.sequence
    outputfile = opts.output
    try:
        generateSequence(inputfile, outputfile, opts.dictionary, timebase)
    except SeqGenException as e:
        print(e.getMsg())
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



####
# json.py:
#
# Encodes GDS objects as JSON.
####
from abc import ABCMeta
from enum import Enum
from inspect import getmembers, isroutine
from typing import Type
from uuid import UUID

import flask.json
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.type_base import BaseType, ValueType

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.cmd_data import CmdData
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.templates.data_template import DataTemplate


def jsonify_base_type(input_type: Type[BaseType]) -> dict:
    """Turn a base type into a JSONable dictionary

    Convert a BaseType (the type, not an instance) into a jsonable dictionary. BaseTypes are converted by reading the
    class properties (without __) and creating the object:

    {
        "name": class name,
        class properties
    }

    Args:
        input_type: input to convert to dictionary
    Return:
        json-able dictionary representing the type
    """
    assert issubclass(input_type, BaseType), "Failure to properly encode data"
    members = getmembers(
        input_type,
        lambda value: not isroutine(value) and not isinstance(value, property),
    )
    jsonable_dict = {name: value for name, value in members if not name.startswith("_")}
    jsonable_dict.update({"name": input_type.__name__})
    return jsonable_dict


def getter_based_json(obj):
    """Converts objects to JSON via get_ methods

    Template functions define a series of get_* methods whose return values need to be serialized. This function
    handles that data.

    Args:
        obj: object to serialize into JSON

    Returns:
        JSON compatible python anonymous type (dictionary)
    """
    anonymous = {}
    getters = [attr for attr in dir(obj) if attr.startswith("get_")]
    for getter in getters:
        # Call the get_ functions, and call all non-static methods
        try:
            func = getattr(obj, getter)
            item = func()
            # If there is a property named "args" it needs to be handled specifically unless an incoming command
            if getter == "get_args":
                args = []
                for arg_spec in item:
                    arg_dict = {
                        "name": arg_spec[0],
                        "description": arg_spec[1] if arg_spec[1] is not None else "",
                        "type": arg_spec[2],
                    }
                    args.append(arg_dict)
                # Fill in our special handling
                item = args
            anonymous[getter.replace("get_", "")] = item
        except TypeError:
            continue
    return anonymous


def minimal_event(obj):
    """Minimal event encoding: time, id, display_text

    Events need time, id, display_text. No other information from the event is necessary for the display.  This will
    minimally encode the data for JSON.

    Args:
        obj: object to serialize into JSON

    Returns:
        JSON compatible python anonymous type (dictionary)
    """
    return {"time": obj.time, "id": obj.id, "display_text": obj.display_text}


def minimal_channel(obj):
    """Minimal channel serialization: time, id, val, and display_text

    Minimally serializes channel values for use with the flask layer. This does away with any unnecessary data by
    serializing only the id, value, and optional display text

    Args:
        obj: object to serialize into JSON

    Returns:
        JSON compatible python anonymous type (dictionary)
    """
    return {
        "time": obj.time,
        "id": obj.id,
        "val": obj.val_obj.val,
        "display_text": obj.display_text,
    }


def minimal_command(obj):
    """Minimal command serialization: time, id, and args values

    Minimally serializes the command values for use with the flask layer. This prevents excess data by keeping the data
    to the minimum instance data for commands including: time, opcode (id), and the value for args.

    Args:
        obj: object to serialize into JSON

    Returns:
        JSON compatible python anonymous type (dictionary)
    """
    return {"time": obj.time, "id": obj.id, "args": obj.get_arg_vals()}


def time_type(obj):
    """Time type serialization

    Serializes the time type into a JSON compatible object.

    Args:
        obj: object to serialize into JSON

    Returns:
        JSON compatible python anonymous type (dictionary)
    """
    assert isinstance(obj, TimeType), "Incorrect type for serialization method"
    return {
        "base": obj.timeBase.value,
        "context": obj.timeContext,
        "seconds": obj.seconds,
        "microseconds": obj.useconds,
    }


def enum_json(obj):
    """Jsonify the python enums!"""
    enum_dict = {"value": str(obj), "values": {}}
    for enum_val in type(obj):
        enum_dict["values"][str(enum_val)] = enum_val.value
    return enum_dict


JSON_ENCODERS = {
    ABCMeta: jsonify_base_type,
    UUID: str,
    ChData: minimal_channel,
    EventData: minimal_event,
    CmdData: minimal_command,
    TimeType: time_type,
}


def default(obj):
    """
    Override the default JSON encoder to pull out a dictionary for our handled types for encoding with the default
    encoder built into flask. This function must convert the given object into a JSON compatable python object (e.g.
    using lists, dictionaries, strings, and primitive types).

    :param obj: obj to encode
    :return: JSON
    """
    if type(obj) in JSON_ENCODERS:
        return JSON_ENCODERS[type(obj)](obj)
    if isinstance(obj, DataTemplate):
        return getter_based_json(obj)
    if isinstance(obj, Enum):
        return enum_json(obj)
    if isinstance(obj, ValueType):
        return obj.val
    return flask.json.provider.DefaultJSONProvider.default(obj)



####
# events.py:
#
# This file captures the HTML endpoint for the events list. This file contains one endpoint which allows for listing
# events after a certain time-stamp.
#
#  GET /events: list events
#      Input Data: {
#                      "start-time": "YYYY-MM-DDTHH:MM:SS.sss" #Start time for event listing
#                  }
####
from fprime_gds.flask.resource import DictionaryResource, HistoryResourceBase


class EventDictionary(DictionaryResource):
    """Channel dictionary shares implementation"""


class EventHistory(HistoryResourceBase):
    """
    Resource supplying the history of events in the system. Includes postprocessing to add in the
    for the display_text attribute which Flask will use.
    """



import argparse
from pathlib import Path

from fprime_gds.common.fpy.parser import parse
from fprime_gds.common.fpy.codegen import compile
from fprime_gds.common.fpy.bytecode.directives import serialize_directives


def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("input", type=Path, help="The input .fpy file")
    arg_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        required=False,
        default=None,
        help="The output .bin path",
    )
    arg_parser.add_argument(
        "-d",
        "--dictionary",
        type=Path,
        required=True,
        help="The FPrime dictionary .json file",
    )

    args = arg_parser.parse_args()

    if not args.input.exists():
        print(f"Input file {args.input} does not exist")
        exit(-1)

    body = parse(args.input.read_text())
    directives = compile(body, args.dictionary)
    output = args.output
    if output is None:
        output = args.input.with_suffix(".bin")
    serialize_directives(directives, output)
    print("Done")


""" fprime_gds.common.zmq_transport:

A set of implementations based on ZeroMQ for data passing amongst components in the fprime_gds. ZeroMQ was chose to
replace the ThreadedTcpServer for several reasons as described below.

1. ZeroMQ is performant and can move 1GB/s/thread
2. ZeroMQ does not require a separate process and may be used w/o tcp sockets
3. ZeroMQ will not reorder packets at high data rates

@author lestarch
"""

import logging
import struct
from typing import Tuple

import zmq

from fprime_gds.common.communication.ground import GroundHandler
from fprime_gds.common.transport import (
    RoutingTag,
    ThreadedTransportClient,
    TransportationException,
)

LOGGER = logging.getLogger("transport")


class ZmqWrapper(object):
    """Handler for ZMQ functions for use in other objects"""

    def __init__(self):
        """Initialize the ZMQ setup"""
        super().__init__()
        self.context = zmq.Context()
        self.zmq_socket_incoming = None
        self.zmq_socket_outgoing = None
        self.pub_topic = None
        self.sub_topic = None
        self.transport_url = None
        self.server = False

    def configure(self, transport_url: Tuple[str], sub_topic: bytes, pub_topic: bytes):
        """Configure the ZeroMQ wrapper

        Configures the ZeroMQ wrapper, but notably does not connect it. This has been separated out because ZeroMQ
        sockets are expected to be created on their governing thread. As such, we just setup the information but defer
        the creation of the sockets until the threads exist.

        Args:
            transport_url: url of the ZeroMQ network to connect to
            sub_topic: subscription topic used to filter incoming messages
            pub_topic: publication topic supplied for remote subscription filters
        """
        assert (
            len(transport_url) == 2
        ), f"Must supply a pair of URLs for ZeroMQ not '{transport_url}'"
        self.pub_topic = pub_topic
        self.sub_topic = sub_topic
        self.transport_url = transport_url

    def make_server(self):
        """Makes this wrapper a server

        ZeroMQ needs at least one server for binding the network to real resources. However, it ZeroMQ doesn't really
        change functionality it provides to the caller. It just changes the call to connect/bind the connection to the
        network.
        """
        assert (
            self.pub_topic is None and self.sub_topic is None
        ), "Cannot make server after connect call"
        self.server = True

    def connect_outgoing(self):
        """Sets up a ZeroMQ connection for outgoing data

        ZeroMQ allows multiple connections to a single endpoint. However, there must be at least one bind client in the
        system. This will connect the ZeroMQ topology and if self.server is set, will bind the outgoing port rather than
        connect it such that this wrapper will act as a host. This only affects outgoing connections as sockets must be
        created on their owning threads.

        The connection is made using self.transport_url, and as such, this must be configured before running. This is
        intended to be called on the sending thread.
        """
        assert (
            self.transport_url is not None and len(self.transport_url) == 2
        ), "Must configure before connecting"
        assert (
            self.zmq_socket_outgoing is None
        ), "Cannot connect outgoing multiple times"
        assert self.pub_topic is not None, "Must configure sockets before connecting"
        self.zmq_socket_outgoing = self.context.socket(zmq.PUB)
        self.zmq_socket_outgoing.setsockopt(zmq.SNDHWM, 0)
        # When set to bind sockets, connect via a bind call
        if self.server:
            server_transport = self.transport_url[1].replace("localhost", "127.0.0.1")
            LOGGER.info("Outgoing binding to: %s", (server_transport))
            self.zmq_socket_outgoing.bind(server_transport)
        else:
            LOGGER.info("Outgoing connecting to: %s", (self.transport_url[0]))
            self.zmq_socket_outgoing.connect(self.transport_url[0])

    def connect_incoming(self):
        """Sets up a ZeroMQ connection for incoming data

        ZeroMQ allows multiple connections to a single endpoint. This only affects incoming connections as sockets must
        be created on their owning threads. This will connect the ZeroMQ topology and if self.server is set, will bind
        the incoming port rather than connect it such that this wrapper will act as a host.

        The connection is made using self.transport_url, and as such, this must be configured before running. This is
        intended to be called on the receiving thread.
        """
        assert (
            self.transport_url is not None and len(self.transport_url) == 2
        ), "Must configure before connecting"
        assert (
            self.zmq_socket_incoming is None
        ), "Cannot connect incoming multiple times"
        assert self.sub_topic is not None, "Must configure sockets before connecting"
        self.zmq_socket_incoming = self.context.socket(zmq.SUB)
        self.zmq_socket_incoming.setsockopt(zmq.RCVHWM, 0)
        self.zmq_socket_incoming.setsockopt(zmq.SUBSCRIBE, self.sub_topic)
        if self.server:
            server_transport = self.transport_url[0].replace("localhost", "127.0.0.1")
            LOGGER.info("Incoming binding to: %s", (server_transport))
            self.zmq_socket_incoming.bind(server_transport)
        else:
            LOGGER.info("Incoming connecting to: %s", (self.transport_url[1]))
            self.zmq_socket_incoming.connect(self.transport_url[1])

    def disconnect_outgoing(self):
        """Disconnect the ZeroMQ sockets"""
        if self.zmq_socket_outgoing is not None:
            self.zmq_socket_outgoing.close()

    def disconnect_incoming(self):
        """Disconnect the ZeroMQ sockets"""
        if self.zmq_socket_incoming is not None:
            self.zmq_socket_incoming.close()

    def terminate(self):
        """Terminate the ZeroMQ context"""
        self.context.term()

    def recv(self, timeout=None):
        """Receive single packet from ZMQ"""
        try:
            self.zmq_socket_incoming.setsockopt(
                zmq.RCVTIMEO, timeout if timeout is not None else -1
            )
            message = self.zmq_socket_incoming.recv()[len(self.sub_topic) :]

        except zmq.Again:
            return b""
        return message

    def send(self, data):
        """Send single message through ZMQ"""
        message_out = self.pub_topic + data
        return self.zmq_socket_outgoing.send(message_out)


class ZmqClient(ThreadedTransportClient):
    """ZeroMQ client to the transport layer

    This client will connect into a ZeroMQ network to act as a data client. Unless the make_server() is called, this
    client will only connect to existing resource and not bind those resources. It uses ThreadedTransportClient to read
    data from the network as part of a thread.

    Note: implementation delegates to a zmq wrapper
    """

    def __init__(self):
        """Create ZMQ wrapper"""
        super().__init__()
        self.zmq = ZmqWrapper()

    def connect(
        self,
        transport_url: Tuple[str],
        sub_routing: RoutingTag,
        pub_routing: RoutingTag,
    ):
        """Connects to the ZeroMQ network"""
        self.zmq.configure(transport_url, sub_routing.value, pub_routing.value)
        self.zmq.connect_outgoing()  # Outgoing socket, for clients, exists on the current thread
        super().connect(transport_url, sub_routing, pub_routing)

    def disconnect(self):
        """Disconnects from ZeroMQ network"""
        self.zmq.disconnect_outgoing()  # Outgoing is on the current thread
        super().disconnect()

    def send(self, data):
        """Send data via ZeroMQ"""
        if data[:4] == b"ZZZZ":
            data = data[4:]
        self.zmq.send(
            data
        )  # Must strip out ZZZZ as that is a ThreadedTcpServer only property

    def recv(self, timeout=None):
        """Receives data from ZeroMQ"""
        return self.zmq.recv(timeout)

    def recv_thread(self):
        """Overrides the recv_thread method

        Overrides the recv_thread method of the superclass such that the ZeroMQ socket may be created/destroyed
        before/after the main recv loop.
        """
        self.zmq.connect_incoming()
        super().recv_thread()  # Contains a while <event> loop, will only return at end of program
        self.zmq.disconnect_incoming()
        self.zmq.terminate()  # Everything should be shutdown and safe to terminate the context


class ZmqGround(GroundHandler):
    """Ground handler implementation for using ZeroMQ as the transport

    This ground handler is built upon the ZmqWrapper subclass an uses zero mq to send data from the communications layer
    to the display and processing layer(s). This effectively acts as the "FSW" side of that interface as it
    frames/deframes packets heading to that layer.

    Since there is likely only one communications client to the FSW users should instantiate with server=True
    to ensure that it binds to resources for the network. This is not forced in case of multiple FSW connections.
    """

    def __init__(self, transport_url, server=True):
        """Initialize this interface with the transport_url needed to connect

        Args:
            transport_url: transport url passed into the zeromq connection
        """
        super().__init__()
        self.zmq = ZmqWrapper()
        self.transport_url = transport_url
        self.timeout = 10
        if server:
            self.zmq.make_server()

    def open(self):
        """Open this ground interface. Delegates to the connect method

        Opens any needed resources and prepares the system for receiving and sending. This means opening the ZeroMQ
        network and setting up routing to incoming as FSW and outgoing to GUI.

        Returns:
            True on successful connection, False on error
        """
        try:
            self.zmq.configure(
                self.transport_url, RoutingTag.FSW.value, RoutingTag.GUI.value
            )
        except TransportationException:
            return False
        return True

    def close(self):
        """Closes the open adapter"""
        # Don't know what else to do, the governing threads are dead, so attempt to close?
        self.zmq.disconnect_incoming()
        self.zmq.disconnect_outgoing()
        self.zmq.terminate()

    def receive_all(self):
        """Receive all available packets

        Receive all packet available from the ground layer. This will return full ground packets up to the uplinker.
        These packets should be fully-deframed and ready for reframing in the comm-layer specified format. With ZeroMQ
        handling whole messages, there is never any outstanding data.

        Returns:
            list deframed packets, outstanding data (always b"")
        """
        if self.zmq.zmq_socket_incoming is None:
            self.zmq.connect_incoming()
        messages = []
        data = None
        while data != b"":
            data = self.zmq.recv(timeout=self.timeout)
            if data != b"":
                # TODO: we need to fix where this is being pulled off, should be done in the framing protocol for uplink
                data = data[
                    4:
                ]  # Strip off the size as this will be re-added by the framing protocol
                messages.append(data)
        return messages

    def send_all(self, frames):
        """Send all the data frames to GUI

        Sends all available frames across the system to the waiting GDS layer. No extra wrapping is performed as ZeroMQ
        handles its own on-wire framing setup.

        Args:
            frames: list of bytes messages to send out
        """
        if self.zmq.zmq_socket_outgoing is None:
            self.zmq.connect_outgoing()
        for packet in frames:
            # TODO: we need to fix where this is being pulled off, should be done in the framing protocol for uplink
            size_bytes = struct.pack(
                ">I", len(packet)
            )  # Add in size bytes as it was stripped in the downlink protocol
            self.zmq.send(size_bytes + packet)



"""
Created on Jan 9, 2015
@author: reder
"""

# Exception classes for all controllers modules


class FprimeGdsException(Exception):
    """
    Base Exception for exceptions that need to be handled at the system-level
    by the GDS
    """

    pass


class GdsDictionaryParsingException(FprimeGdsException):
    pass


# Note: Gse is the historical name for what is now called the GDS
class GseControllerException(FprimeGdsException):
    def __init__(self, val):
        self.except_msg = val
        super().__init__(val)

    def getMsg(self):
        return self.except_msg


class GseControllerUndefinedDirectoryException(GseControllerException):
    def __init__(self, val):
        super().__init__(f"Path does not exist: {str(val)}!")


class GseControllerUndefinedFileException(GseControllerException):
    def __init__(self, val):
        super().__init__(f"Path does not exist: {str(val)}!")


class GseControllerParsingException(GseControllerException):
    def __init__(self, val):
        super().__init__(f"Parsing error: {str(val)}")
